﻿using FirebirdSql.Data.FirebirdClient;
using FirebirdSql.Data.Services;
using FirebirdSql.Data.Types;
using Microsoft.VisualBasic;
using System;
using System.ComponentModel;
using System.Data;
using System.Data.Common;
using System.Diagnostics;
using System.Reflection;
using IniParser;
using IniParser.Model;

namespace firebird_nbackup_crashes_server
{
    internal class Program
    {
        static void Main(string[] args)
        {
            Console.WriteLine("[Start]");

            var parser = new FileIniDataParser();

            if (!File.Exists("config.ini"))
            {
                Console.WriteLine("config.ini missing");
                return;
            }
            IniData data = parser.ReadFile("config.ini");

            var cb = new FbConnectionStringBuilder();
            cb.DataSource = "127.0.0.1";
            cb.Port = 3054;
            cb.Password = "123";
            cb.UserID = "SYSDBA";
            cb.Database = System.IO.Path.Combine(AppContext.BaseDirectory, "test.fdb");
            cb.Pooling = false;
            cb.WireCrypt = FbWireCrypt.Enabled;
            cb.Compression = false;

            cb.Port = int.Parse(data["firebird"]["Port"]);
            cb.UserID = data["firebird"]["UserID"];
            cb.Password = data["firebird"]["Password"];

            int N = 20;

            for (int i = 0; i < N; i++)
            {
                cb.Database = System.IO.Path.Combine(AppContext.BaseDirectory, $"db-{i}.fdb");
                var constring = cb.ToString();

                Console.WriteLine($"CreateDb {i}");

                if (System.IO.File.Exists(cb.Database))
                {
                    System.IO.File.Delete(cb.Database);
                }

                if (!System.IO.File.Exists(cb.Database))
                {
                    FbConnection.CreateDatabase(constring, 8192, true, true);
                }
            }

            var alltasks = new List<Task>();
            for (int i = 0; i < N; i++)
            {
                var dbfn = System.IO.Path.Combine(AppContext.BaseDirectory, $"db-{i}.fdb");
                cb.Database = dbfn;
                var constring = cb.ToString();

                var index = i;

                var t = new Task<bool>(() =>
                {
                    BackupDb(constring, index, false);
                    BackupDb(constring, index, true);
                    return true;
                });
                alltasks.Add(t);

            }

            foreach (var t in alltasks)
            {
                t.Start();
            }

            foreach (var t in alltasks)
            {
                t.Wait();
            }

            Console.WriteLine("[Complete]");
        }

        private static void EnsureDatabaseNotInBackupMode(string constring)
        {
            using var con = new FbConnection(constring);
            con.Open();            
            using var tran = con.BeginTransaction();            
            using var cmd = con.CreateCommand();
            cmd.Transaction = tran;
            cmd.CommandText = "ALTER DATABASE END BACKUP";

            try
            {
                cmd.ExecuteNonQuery();
                tran.Commit();
            }
            catch (Exception)
            {
                tran.Rollback(); //already not in backup mode
            }

            con.Close();
        }

        private static void BackupDb(string constring, int index, bool IncrementalDeltaBackup)
        {
            try
            {
                Console.WriteLine($"BackupDb {index} IncrementalDeltaBackup:{IncrementalDeltaBackup}");

                EnsureDatabaseNotInBackupMode(constring);

                string bkpfn_full = System.IO.Path.Combine(AppContext.BaseDirectory, $"db-{index}.full.nbkp");
                if (IncrementalDeltaBackup)
                {
                    bkpfn_full = System.IO.Path.Combine(AppContext.BaseDirectory, $"db-{index}.delta.nbkp");
                }

                string bkpfntemp = bkpfn_full;


                if (System.IO.File.Exists(bkpfntemp))
                {
                    try
                    {
                        System.IO.File.Delete(bkpfntemp);
                    }
                    catch (Exception exd)
                    {
                        Console.WriteLine($"{exd.Message}");
                    }

                }

                var bkp = new FbNBackup();
                bkp.BackupFile = bkpfntemp;
                bkp.ConnectionString = constring;
                bkp.Options = FbNBackupFlags.NoDatabaseTriggers;
                if (IncrementalDeltaBackup)
                {
                    bkp.Level = 1;
                }
                else
                {
                    bkp.Level = 0;
                }
                var backup_output = new List<string>();
                // so that program blocks till complete
                bkp.ServiceOutput += (sender, e) =>
                {
                    Console.WriteLine($"Backup:{e.Message}");
                };
                bkp.InfoMessage += (sender, e) =>
                {
                    Console.WriteLine($"INFO) {e.Message}");
                    foreach (var err in e.Errors)
                    {
                        Console.WriteLine($"INFO) Class:{err.Class} LineNumber:{err.LineNumber} Number:{err.Number} Message:{err.Message}");
                    }
                };
                Console.WriteLine($"Starting Backup...");

                bkp.Execute();
                Console.WriteLine($"Backup Done");
            }
            catch (Exception ex)
            {
                Console.WriteLine($"Backup FAIL");
                Console.WriteLine(ex.ToString());
                Console.WriteLine();
            }

        }

    }
}